/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.
 */

#include "config/configurationcommandline.h"

#include <logfunc.hpp>

#include "boost_assert_handler.h"
#include "config/configurationoptions.h"
#include "config/option/hostnameoption.h"

using namespace std;
using boost::program_options::command_line_parser;
using boost::program_options::options_description;
using boost::program_options::parsed_options;
using boost::program_options::positional_options_description;
using boost::program_options::variables_map;
using I2n::Logger::GlobalLogger;

//-----------------------------------------------------------------------------
// ConfigurationCommandLine
//-----------------------------------------------------------------------------

/**
 * @brief Parameterized constructor. Never called since is private
 *
 * @param _argc The number of arguments in command line.
 * @param _argv A vector containing the command line elements.
 */
ConfigurationCommandLine::ConfigurationCommandLine(
        const int _argc,
        const char **_argv
) :
    argc( _argc ),
    argv( _argv ) // TODO perform a deep copy, not swallow
{
    BOOST_ASSERT( argc > 0 );
    BOOST_ASSERT( argv != NULL );
}

/**
 * @brief Copy constructor. Never called since is private
 *
 * @param other A object to be copied.
 */
ConfigurationCommandLine::ConfigurationCommandLine(
        const ConfigurationCommandLine &other
) :
    argc( other.argc ),
    argv( other.argv ) // TODO perform a deep copy, not swallow
{
}

/**
 * @brief Destructor.
 */
ConfigurationCommandLine::~ConfigurationCommandLine()
{
}

ConfigurationCommandLine & ConfigurationCommandLine::operator =(
        const ConfigurationCommandLine &other
)
{
    argc = other.argc;
    argv = other.argv; // TODO perform a deep copy, not swallow

    return *this;
}

/**
 * @brief Reads the command line, storing the parsed tokens in the
 * @c variables_map.
 *
 * @param vm The variables_map object to store configuration tokens.
 *
 * @return @c true on success and @c false otherwise.
 */
bool ConfigurationCommandLine::process( variables_map *vm )
{
    BOOST_ASSERT( vm != NULL );

    try
    {
        ConfigurationOptions options;
        HostNameOption host_name_option;

        options_description generic = options.get_generic_options();
        options_description config = options.get_configuration_options();

        options_description cmdline_options;
        cmdline_options.add( generic ).add( config );

        positional_options_description p;
        (void) p.add( host_name_option.get_command_string().c_str(), -1 );

        parsed_options parsed_opt = command_line_parser( argc, const_cast<char**>( argv ) ).
                options( cmdline_options ).
                positional( p ).
                run();
        store( parsed_opt, *vm );
        notify( *vm );
    }
    catch ( const std::exception &ex )
    {
        GlobalLogger.error() << "Error parsing command line options: " << ex.what() << endl;
        return false;
    }

    return true;
}

/**
 * @brief Store the parsed tokens in @a configuration.
 *
 * @param vm The variables_map object where the configuration tokens are stored.
 * @param configuration A @c Configuration object to where to save the
 * configuration.
 *
 * @return @c true if parsed successfully, or @c false otherwise.
 */
bool ConfigurationCommandLine::parse(
        const variables_map &vm,
        Configuration *configuration
)
{
    BOOST_ASSERT( configuration != NULL );

    ConfigurationOptions options;

    bool generic_parsed = options.parse_generic_options( vm, configuration );
    // Do not proceed if it is one of the halt options (i.e. --version, --help)
    bool halt = options.halt_on_generic_options( vm );
    if ( halt )
    {
        return false;
    }
    bool configuration_parsed = options.parse_configuration_options( vm, configuration );

    return ( generic_parsed || configuration_parsed );
}
